(* interface to UFTP client(s) *)
(* solyga@gmx.de, 2019-03-15 *)
IMPLEMENTATION MODULE CliIO;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT WinSock;
IMPORT IO;
IMPORT Lib2, Lib3, SO, Utils, Input, Config, Tmr, Tmrs;

CONST
  period = 20;  (* ms *)

TYPE
  TSocket  = WinSock.SOCKET;
  TSockSet = WinSock.fd_set;
  TAdrDesc = WinSock.sockaddr_in;

CONST
  adrNull = 0;  (* = WinSock.INADDR_ANY *)
  prtNull = 0;  (* yields ephemeral port *)
  socNull = WinSock.INVALID_SOCKET;  (* -1 *)

TYPE
  TCntDesc = RECORD
    txn,
    txe: CARDINAL;
    rxn,
    rxe: CARDINAL;
  END;

VAR
  hnd: THandler;
  soc: TSocket;
  cnt: TCntDesc;
  tcs: Tmr.TTicks;


PROCEDURE Message( s-: ARRAY OF CHAR );
BEGIN
  IO.WrStr( 'CliIO.' );
  IO.WrStr( s );
  IO.WrLn();
END Message;

PROCEDURE Error( s-: ARRAY OF CHAR );
BEGIN
  Message( s );
END Error;

PROCEDURE Fatal( s-: ARRAY OF CHAR );
BEGIN
  Message( s );
  HALT;
END Fatal;


PROCEDURE Trigger();
VAR
  now: Tmr.TTicks;
BEGIN
  Input.Trigger();
  now := Tmr.Ticks();
  Tmrs.Trigger( Tmr.Diff( tcs, now ) );
  tcs := now;
END Trigger;


PROCEDURE NewSock(): TSocket;
VAR
  s: TSocket;
  b: SO.TBuffer;
  t: ARRAY [0..63] OF CHAR;
BEGIN
  s := WinSock.socket( WinSock.AF_INET, WinSock.SOCK_DGRAM, WinSock.IPPROTO_UDP );
  IF s = WinSock.INVALID_SOCKET THEN
    SO.Init( b, t );
    SO.WrStr( b, 'NewSock: socket() failed, err ' );
    SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
    Fatal( t );
  END;
  RETURN s;
END NewSock;

PROCEDURE BindSock( s: TSocket; la: TAddress; lp: TPort );
VAR
  a: TAdrDesc;
  b: SO.TBuffer;
  t: ARRAY [0..63] OF CHAR;
BEGIN
  a.sin_family      := WinSock.AF_INET;
  a.sin_addr.s_addr := la;
  a.sin_port        := WinSock.htons( lp );
  IF WinSock.bind( s, ADR(a), SIZE(a) ) = WinSock.SOCKET_ERROR THEN
    SO.Init( b, t );
    SO.WrStr( b, 'BindSock: bind() for ' );
    Utils.WrIP( b, la );
    SO.WrChar( b, ':' );
    SO.WrCard( b, lp, 1 );
    SO.WrStr( b, ' failed, err ' );
    SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
    Fatal( t );
  END;
END BindSock;

PROCEDURE GetSockAdr( s: TSocket; VAR la: TAddress; VAR lp: TPort );
VAR
  a: TAdrDesc;
  l: INTEGER;
  b: SO.TBuffer;
  t: ARRAY [0..63] OF CHAR;
BEGIN
  Lib2.Fill( a, 0 );
  l := SIZE( a );
  IF WinSock.getsockname( s, ADR(a), l ) = WinSock.SOCKET_ERROR THEN
    SO.Init( b, t );
    SO.WrStr( b, 'GetSockAdr: getsockname() failed, err ' );
    SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
    Fatal( t );
  END;
  IF a.sin_family # WinSock.AF_INET THEN
    SO.Init( b, t );
    SO.WrStr( b, 'GetSockAdr: getsockname() returned invalid family value' );
    Fatal( t );
  END;
  IF l >= 8 THEN
    la := a.sin_addr.s_addr;
    lp := WinSock.ntohs( a.sin_port );
  ELSE
    SO.Init( b, t );
    SO.WrStr( b, 'GetSockAdr: getsockname() returned short address record' );
    Error( t );
  END;
END GetSockAdr;

PROCEDURE WrSock( s: TSocket; ra: TAddress; rp: TPort; x-: ARRAY OF BYTE; l: CARDINAL ): BOOLEAN;
VAR
  a: TAdrDesc;
  r: INTEGER;
BEGIN
  Lib2.Fill( a, 0 );
  a.sin_family      := WinSock.AF_INET;
  a.sin_port        := WinSock.htons( rp );
  a.sin_addr.s_addr := ra;
  r := WinSock.sendto( s, ADR(x), INTEGER(l), 0, ADR(a), SIZE(a) );
  RETURN r = INTEGER( l );
END WrSock;

PROCEDURE RdSock( s: TSocket; VAR ra: TAddress; VAR rp: TPort; VAR x: ARRAY OF BYTE; VAR l: CARDINAL ): BOOLEAN;
VAR
  a: TAdrDesc;
  al: INTEGER;
  r: INTEGER;
BEGIN
  Lib2.Fill( a, 0 );
  al := SIZE( a );
  r := WinSock.recvfrom( s, ADR(x), SIZE(x), 0, ADR(a), al );
  (* buffer too small  =>  r = -1, WSAEMSGSIZE, excess data lost *)
  IF ( r < 0 ) OR ( al < 8 ) THEN
    ra := 0;
    rp := 0;
    l := 0;
    RETURN FALSE;
  END;
  ra := a.sin_addr.s_addr;
  rp := WinSock.ntohs( a.sin_port );
  l := r;
  RETURN TRUE;
END RdSock;


PROCEDURE IncCnt( VAR x: CARDINAL );
BEGIN
  IF x < MAX( CARDINAL ) THEN INC( x ) ELSE x := 0 END;
END IncCnt;


PROCEDURE Install( h: THandler );
BEGIN
  IF hnd # NULLPROC THEN Fatal( 'Install: Called twice' ); END;
  hnd := h;
END Install;

PROCEDURE Send( d-: TDstDesc; x-: ARRAY OF BYTE; l: CARDINAL );
VAR
  ok: BOOLEAN;

  PROCEDURE Show();
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Send: ' );
    SO.WrCard( b, l, 1 );
    SO.WrStr( b, ' bytes sent to ' );
    Utils.WrIP( b, d.a );
    SO.WrChar( b, ':' );
    SO.WrCard( b, d.p, 1 );
    IF ok THEN SO.WrStr( b, ' ok' ) ELSE SO.WrStr( b, ' nok' ) END;;
    Message( s );
  END Show;

BEGIN
  ok := WrSock( soc, d.a, d.p, x, l );
  Show();
  IF ok THEN
    IncCnt( cnt.txn );
  ELSE
    IncCnt( cnt.txe );
  END;
END Send;


PROCEDURE Receive();
VAR
  r: TRxInfDesc;
  x: ARRAY [1..lenMax] OF BYTE;
  l: CARDINAL;
  ok: BOOLEAN;

  PROCEDURE Show();
  CONST
    ena = FALSE;
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'Receive: ' );
    SO.WrCard( b, l, 1 );
    SO.WrStr( b, ' bytes received from ' );
    Utils.WrIP( b, r.d.a );
    SO.WrChar( b, ':' );
    SO.WrCard( b, r.d.p, 1 );
    IF ok THEN SO.WrStr( b, ' ok' ) ELSE SO.WrStr( b, ' nok' ) END;;
    Message( s );
  END Show;

BEGIN
  ok := RdSock( soc, r.d.a, r.d.p, x, l );
  Show();
  IF ok THEN
    IncCnt( cnt.rxn );
    IF hnd # NULLPROC THEN hnd( r, x, l ) END;
  ELSE
    IncCnt( cnt.rxe );
  END;
END Receive;

PROCEDURE Execute();
CONST
  to = WinSock.TIMEVAL{ period DIV 1000, ( period MOD 1000 ) * 1000 };
VAR
  s: TSockSet;
  r: INTEGER;
  b: SO.TBuffer;
  t: ARRAY [0..63] OF CHAR;
BEGIN
  LOOP
    Trigger();
    WinSock.FD_ZERO( s );
    WinSock.FD_SET( soc, s );
    r := WinSock.select( 0, s, NIL, NIL, ADR(to) );
    IF r = WinSock.SOCKET_ERROR THEN
      SO.Init( b, t );
      SO.WrStr( b, 'Execute: select() failed, err ' );
      SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
      Fatal( t );
    ELSIF r > 0 THEN
      (* r = count of ready sockets *)
      Receive();
    ELSE
      (* timeout *)
    END;
  END;
  SO.Init( b, t );
  SO.WrStr( b, 'Execute: RETURN' );
  Fatal( t );
END Execute;


PROCEDURE InitWSA();
VAR
  w: WinSock.WSADATA;
  e: INTEGER;
  b: SO.TBuffer;
  s: ARRAY [0..63] OF CHAR;

  PROCEDURE Show();
  CONST
    ena = FALSE;
  BEGIN
    IF ~ ena THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'InitWSA: "' );
    SO.WrStr( b, w.szDescription );
    SO.WrStr( b, '", ver ' );
    b.c.p := '0';
    SO.WrHex( b, w.wVersion, 4 );
    Message( s );
  END Show;

BEGIN
  e := WinSock.WSAStartup( 0202H, w );
  IF e # 0 THEN
    SO.Init( b, s );
    SO.WrStr( b, 'InitWSA: WSAStartup() failed, err ' );
    SO.WrLngInt( b, WinSock.WSAGetLastError(), 1 );
    Fatal( s );
  END;
  Show();
END InitWSA;

PROCEDURE InitSock();
VAR
  b: SO.TBuffer;
  s: ARRAY [0..63] OF CHAR;
  d: TDstDesc;
BEGIN
  soc := NewSock();
  BindSock( soc, Config.data.adr, Config.data.prt );
  GetSockAdr( soc, d.a, d.p );
  IF Config.data.vrb THEN
    SO.Init( b, s );
    SO.WrStr( b, 'InitSock: Listening on ' );
    Utils.WrIP( b, d.a );
    SO.WrChar( b, ':' );
    SO.WrCard( b, d.p, 1 );
    Message( s );
  END;
END InitSock;

BEGIN
  hnd := NULLPROC;
  soc := socNull;
  Lib2.Fill( cnt, 0 );
  tcs := Tmr.Ticks();

  InitWSA();
  InitSock();
END CliIO.
